"""Test tools enabling compatibility with QCElemental"""

import numpy as np
from qcelemental.models import AtomicInput, AtomicResult

# from qcio import ProgramFailure
from qcio.qcel import from_qcel_output_results, to_qcel_input


def test_input_to_qcel(prog_input_factory):
    """Test that input converts to qcel input"""
    pi_energy = prog_input_factory("energy")
    qcel_input_dict = to_qcel_input(pi_energy)
    qcel_atomic_input = AtomicInput(**qcel_input_dict)

    assert qcel_atomic_input.molecule.symbols.tolist() == pi_energy.structure.symbols
    assert np.array_equal(
        qcel_atomic_input.molecule.geometry, pi_energy.structure.geometry
    )
    assert qcel_atomic_input.molecule.molecular_charge == pi_energy.structure.charge
    assert qcel_atomic_input.model.method == pi_energy.model.method
    assert qcel_atomic_input.model.basis == pi_energy.model.basis
    assert qcel_atomic_input.driver == "energy"
    assert qcel_atomic_input.keywords == pi_energy.keywords
    assert qcel_atomic_input.extras == pi_energy.extras


def test_input_to_output_from_qcel_output_with_wfn(prog_input_factory):
    """Test that input converts to qcel input"""
    pi_energy = prog_input_factory("energy")

    # Check that the input converts to a qcel input
    qcel_atomic_input = AtomicInput(**to_qcel_input(pi_energy))

    qcel_atomic_output = AtomicResult(
        **{
            **qcel_atomic_input.dict(),
            **{
                "protocols": {"wavefunction": "all"},
                "properties": {
                    "nuclear_repulsion_energy": 1.0,
                    "return_energy": 1.0,
                    "calcinfo_natom": 3,
                },
                "return_result": 1.0,
                "success": True,
                "wavefunction": {
                    "scf_eigenvalues_a": [1, 2, 3, 4, 5, 6],
                    "restricted": True,
                    "basis": {
                        "name": "6-31g",
                        "center_data": {},
                        "atom_map": [],
                    },
                },
            },
        },
    )
    results = from_qcel_output_results(qcel_atomic_output.dict())

    assert results.energy == qcel_atomic_output.return_result
    assert (
        results.nuclear_repulsion_energy
        == qcel_atomic_output.properties.nuclear_repulsion_energy
    )
    assert results.calcinfo_natoms == qcel_atomic_output.properties.calcinfo_natom


def test_input_to_output_from_qcel_output_no_wfn(prog_input_factory):
    """Test that input converts to qcel input"""
    calc_input_energy = prog_input_factory("energy")
    qcel_atomic_input = AtomicInput(**to_qcel_input(calc_input_energy))

    qcel_atomic_output = AtomicResult(
        **{
            **qcel_atomic_input.dict(),
            **{
                "properties": {
                    "nuclear_repulsion_energy": 1.0,
                    "return_energy": 1.0,
                    "calcinfo_natom": 3,
                },
                "return_result": 1.0,
                "success": True,
            },
        },
    )
    results = from_qcel_output_results(qcel_atomic_output.dict())

    assert results.energy == qcel_atomic_output.return_result
    assert (
        results.nuclear_repulsion_energy
        == qcel_atomic_output.properties.nuclear_repulsion_energy
    )
    assert results.calcinfo_natoms == qcel_atomic_output.properties.calcinfo_natom



import pickle

import numpy as np
import pytest
from pydantic import ValidationError

from qcio import (
    FileInput,
    Files,
    OptimizationData,
    ProgramInput,
    Provenance,
    Results,
    SinglePointData,
    Structure,
    Wavefunction,
)


def test_gradient_converted_np_array():
    """Test that SinglePointData converts gradient to np array"""
    gradient = [1.0, 2.0, 3.0, 4.0, 5.0, 6.0]
    data = SinglePointData(gradient=gradient)
    assert isinstance(data.gradient, np.ndarray)
    assert data.gradient.dtype == np.float64


def test_hessian_converted_np_array():
    """Test that SinglePointData converts hessian to np array"""
    hessian = [float(i) for i in range(9)]
    data = SinglePointData(hessian=hessian)
    assert isinstance(data.hessian, np.ndarray)
    assert data.hessian.dtype == np.float64


def test_single_point_casts_gradient_to_n_by_3(prog_input_factory):
    """Test that SinglePointData casts gradient to n by 3"""
    pi_gradient = prog_input_factory("gradient")
    n_atoms = len(pi_gradient.structure.symbols)
    gradient = [float(i) for i in range(n_atoms * 3)]
    data = SinglePointData(gradient=gradient)
    assert data.gradient.shape == (n_atoms, 3)
    # Assert that gradient is numpy array of floats
    assert data.gradient.dtype == np.float64


def test_single_point_success_casts_hessian_to_3n_by_3n(prog_input_factory):
    """Test that SinglePointData casts hessian to 3n x 3n"""
    pi_hessian = prog_input_factory("hessian")
    n_atoms = len(pi_hessian.structure.symbols)
    hessian = [float(i) for i in range(n_atoms**2 * 3**2)]
    data = SinglePointData(hessian=hessian)
    assert data.hessian.shape == (n_atoms * 3, n_atoms * 3)
    # Assert that hessian is numpy array of floats
    assert data.hessian.dtype == np.float64


def test_single_point_data_normal_modes_cartesian_shape(prog_input_factory):
    """Test that SinglePointData normal_modes_cartesian are n_modes x n_atoms x 3"""
    pi_energy = prog_input_factory("energy")
    n_atoms = len(pi_energy.structure.symbols)
    n_atoms * 3
    data = SinglePointData(
        energy=-1.0,
        freqs_wavenumber=[1.0, 2.0, 3.0],  # Given so there are 3 normal modes
        # fmt: off
        normal_modes_cartesian=np.array(
            [
                -1.45413605e-07,
                7.31568094e-02,
                3.49777695e-34,
                -4.00280118e-01,
                -5.80603129e-01,
                8.72597124e-50,
                4.00282426e-01,
                -5.80601321e-01,
                -2.61779137e-49,
                -8.40792347e-07,
                4.64736028e-02,
                -7.85031947e-33,
                6.02403552e-01,
                -3.68838662e-01,
                -1.08372224e-49,
                -6.02390207e-01,
                -3.68828204e-01,
                3.25116671e-49,
                -6.99575075e-02,
                -7.31659905e-07,
                9.73449382e-38,
                5.55204565e-01,
                -4.35072838e-01,
                1.17356863e-54,
                5.55217962e-01,
                4.35084451e-01,
                -3.52070589e-54,
            ]
        ),
        # fmt: on
    )
    assert data.normal_modes_cartesian.shape == (3, 3, 3)


def test_return_result(prog_input_factory):
    """Test that return_result returns the requested result"""
    # TODO: Remove this test once I depreciate .return_result
    calc_input_energy = prog_input_factory("energy")
    energy = 1.0
    n_atoms = len(calc_input_energy.structure.symbols)
    gradient = np.arange(n_atoms * 3).reshape(n_atoms, 3)
    hessian = np.arange(n_atoms**2 * 3**2).reshape(n_atoms * 3, n_atoms * 3)

    results = Results(
        input_data=calc_input_energy,
        success=True,
        data={
            "energy": energy,
            "gradient": gradient,
            "hessian": hessian,
        },
        provenance={"program": "qcio-test-suite"},
    )
    assert results.return_result == energy
    assert results.return_result == results.data.energy

    pi_gradient = prog_input_factory("gradient")
    results = Results(**{**results.model_dump(), **{"input_data": pi_gradient}})  # noqa: E501
    assert np.array_equal(results.return_result, gradient)
    assert np.array_equal(results.return_result, results.data.gradient)

    pi_hessian = prog_input_factory("hessian")
    results = Results(**{**results.model_dump(), **{"input_data": pi_hessian}})

    assert np.array_equal(results.return_result, hessian)
    assert np.array_equal(results.return_result, results.data.hessian)


def test_successful_result_serialization(results):
    """Test that successful result serializes and deserializes"""
    serialized = results.model_dump_json()
    deserialized = Results.model_validate_json(serialized)
    assert deserialized == results
    assert deserialized.data == results.data
    assert deserialized.input_data == results.input_data
    assert deserialized.provenance.program == "qcio-test-suite"
    assert deserialized.logs == results.logs
    assert deserialized.extras == results.extras
    assert deserialized.return_result == results.return_result
    assert deserialized.data.energy == results.data.energy
    assert np.array_equal(deserialized.data.gradient, results.data.gradient)
    assert np.array_equal(deserialized.data.hessian, results.data.hessian)


def test_wavefunction_to_numpy():
    """Test that wavefunction converts to numpy array"""
    wavefunction = Wavefunction(
        scf_eigenvalues_a=[1.0, 2.0, 3.0, 4.0, 5.0, 6.0],
        scf_eigenvalues_b=[1.0, 2.0, 3.0, 4.0, 5.0, 6.0],
        scf_occupations_a=[1.0, 2.0, 3.0, 4.0, 5.0, 6.0],
        scf_occupations_b=[1.0, 2.0, 3.0, 4.0, 5.0, 6.0],
    )
    assert isinstance(wavefunction.scf_eigenvalues_a, np.ndarray)
    assert isinstance(wavefunction.scf_eigenvalues_b, np.ndarray)
    assert isinstance(wavefunction.scf_occupations_a, np.ndarray)
    assert isinstance(wavefunction.scf_occupations_b, np.ndarray)
    assert wavefunction.scf_eigenvalues_a.dtype == np.float64
    assert wavefunction.scf_eigenvalues_b.dtype == np.float64
    assert wavefunction.scf_occupations_a.dtype == np.float64
    assert wavefunction.scf_occupations_b.dtype == np.float64


def test_correct_generic_instantiates_and_equality_checks_pass(results, tmp_path):
    """
    This test checks the Results.model_post_init method to ensure that the
    correct generic types are instantiated and that equality checks pass. It also
    tests the serialization and deserialization of the Results class with and
    without types. The point is that end users should be able to ignore generics
    specification for common tasks like opening files and the correct types will be
    set for them.
    """
    results_dict = results.model_dump()
    ## SinglePointResults
    # Do not pass types
    wo_types = Results(**results_dict)
    # Pass types
    w_types = Results[ProgramInput, SinglePointData](**results_dict)

    ## OptimizationResults
    results_dict["input_data"]["calctype"] = "optimization"
    results_dict["data"] = OptimizationData(trajectory=[wo_types])
    # Do not pass types
    wo_types_opt = Results(**results_dict)
    # Pass types
    w_types_opt = Results[ProgramInput, OptimizationData](**results_dict)

    # Serialize and Save
    wo_types.save(tmp_path / "out.json")
    w_types_opt.save(tmp_path / "opt.json")

    # Reopen with or without types
    wo_types_opened = Results.open(tmp_path / "out.json")
    w_types_opened = Results[ProgramInput, SinglePointData].open(tmp_path / "out.json")

    # Reopen with or without types
    wo_types_opened_opt = Results.open(tmp_path / "opt.json")
    w_types_opened_opt = Results[ProgramInput, OptimizationData].open(
        tmp_path / "opt.json"
    )

    # Make sure all objects have the same class (with correct types) and values
    assert wo_types == w_types == wo_types_opened == w_types_opened
    assert wo_types_opt == w_types_opt == wo_types_opened_opt == w_types_opened_opt


def test_non_file_success_always_has_result(prog_input_factory):
    pi_energy = prog_input_factory("energy")
    with pytest.raises(ValidationError):
        Results[ProgramInput, SinglePointData](
            success=True,
            input_data=pi_energy,
            logs="program standard out...",
            data=None,
            provenance={"program": "qcio-test-suite"},
        )


def test_primary_result_must_be_present_on_success(results):
    for calctype in ["energy", "gradient", "hessian"]:
        po_dict = results.model_dump()
        po_dict["input_data"]["calctype"] = calctype
        po_dict["data"][calctype] = None
        with pytest.raises(ValidationError):
            Results[ProgramInput, SinglePointData](**po_dict)


@pytest.mark.parametrize(
    "input_data, data, success, expected_input_type, expected_result_type",
    [
        pytest.param(
            "file_input",
            Files(),
            True,
            FileInput,
            Files,
            id="FileInput-Files-Success",
        ),
        pytest.param(
            "file_input",
            Files(),
            False,
            FileInput,
            Files,
            id="FileInput-Files-Failure",
        ),
    ],
    indirect=["input_data"],
)
def test_pickle_serialization_of_program_output_parametrized(
    input_data,
    data,
    success,
    expected_input_type,
    expected_result_type,
    request,
):
    """This test checks that all the dynamic types are correctly set when
    deserializing a Results object. It also checks that the equality
    checks pass for the deserialized object. It also ensure that the types
    set can be looked up by pickle in qcio.models.outputs.Results.s
    """

    provenance = Provenance(program="qcio-test-suite")
    traceback = None
    if success is False:
        traceback = "Fake traceback"

    # Generics specified at instantiation of Results
    prog_output = Results[type(input_data), type(data)](
        input_data=input_data,
        data=data,
        success=success,
        provenance=provenance,
        traceback=traceback,
    )
    serialized = pickle.dumps(prog_output)
    deserialized = pickle.loads(serialized)
    assert deserialized == prog_output

    # Generics not specified at instantiation of Results
    prog_output = Results(
        input_data=input_data,
        data=data,
        success=success,
        provenance=provenance,
        traceback=traceback,
    )
    serialized = pickle.dumps(prog_output)
    deserialized = pickle.loads(serialized)
    assert deserialized == prog_output

    # Generics not specified at instantiation of Results
    unspecified_po = Results(**prog_output.model_dump())
    serialized = pickle.dumps(prog_output)
    deserialized = pickle.loads(serialized)
    assert deserialized == unspecified_po

    # Generics specified and .results is not passed
    prog_output_dict = prog_output.model_dump()
    prog_output_dict.update({"success": False, "traceback": "Traceback: ..."})
    no_results = Results(**prog_output_dict)
    serialized = pickle.dumps(no_results)
    deserialized = pickle.loads(serialized)
    assert deserialized == no_results

    # Dynamically specified generics
    dynamic_generics = Results[type(prog_output.input_data), type(prog_output.data)](
        **prog_output.model_dump()
    )
    serialized = pickle.dumps(dynamic_generics)
    deserialized = pickle.loads(serialized)
    assert deserialized == dynamic_generics


def test_pickle_serialization_of_program_output():
    prog_output = Results[ProgramInput, SinglePointData](
        input_data=ProgramInput(
            structure=Structure(
                symbols=["O", "H", "H"],
                geometry=np.array([0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0, 1.0, 0.0]),
                charge=0,
                multiplicity=1,
                connectivity=[(0, 1, 1.0), (0, 2, 1.0)],
            ),
            calctype="energy",
            model={"method": "hf", "basis": "sto-3g"},
            keywords={
                "maxiter": 100,
                "purify": "no",
                "some-bool": False,
                "displacement": 1e-3,
                "thermo_temp": 298.15,
            },
        ),
        success=True,
        # traceback="Traceback: ...",
        logs="program standard out...",
        data=SinglePointData(
            energy=1.0,
            extras={"some_extra_result": 1},
        ),
        provenance={"program": "qcio-test-suite", "scratch_dir": "/tmp/qcio"},
        extras={"some_extra": 1},
    )
    # Generics specified at instantiation of Results
    serialized = pickle.dumps(prog_output)
    deserialized = pickle.loads(serialized)
    assert deserialized == prog_output

    # Generics not specified at instantiation of Results
    unspecified_po = Results(**prog_output.model_dump())
    serialized = pickle.dumps(prog_output)
    deserialized = pickle.loads(serialized)
    assert deserialized == unspecified_po

    # Generics specified and .data is None
    prog_output_dict = prog_output.model_dump()
    prog_output_dict.update(
        {"data": Files(), "success": False, "traceback": "Traceback: ..."}
    )
    # Fails
    no_data = Results(**prog_output_dict)
    serialized = pickle.dumps(no_data)
    deserialized = pickle.loads(serialized)
    assert deserialized == no_data

    # Dynamically specified generics
    dynamic_generics = Results[type(prog_output.input_data), type(prog_output.data)](
        **prog_output.model_dump()
    )
    serialized = pickle.dumps(dynamic_generics)
    deserialized = pickle.loads(serialized)
    assert deserialized == dynamic_generics


def test_ensure_result_present_on_single_point_data_validator():
    with pytest.raises(ValidationError):
        SinglePointData()


def test_compatibility_layer_for_files_on_results(prog_input_factory):
    """Test that the compatibility layer for files on Results works"""
    energy_input = prog_input_factory("energy")
    # Passing Files as a dictionary
    files = {"file1": "file1.txt", "file2": "file2.txt"}

    po = Results(
        input_data=energy_input,
        success=True,
        data=SinglePointData(energy=-1.0),
        provenance={"program": "qcio-test-suite"},
        files=files,
    )
    assert po.data.files == files


def test_compatibility_layer_for_noresults_prog_outputs(test_data_dir):
    """Ensure old Results with NoResults can still be loaded."""
    Results.model_validate_json((test_data_dir / "po_noresults.json").read_text())


def test_compatibility_layer_for_results_on_program_output(prog_input_factory):
    """Test that the compatibility layer for files on Results works"""
    energy_input = prog_input_factory("energy")
    # Passing Files as a dictionary
    files = {"file1": "file1.txt", "file2": "file2.txt"}
    results_dict = {
        "input_data": energy_input,
        "success": True,
        "results": {"energy": -1.0},
        "provenance": {"program": "qcio-test-suite"},
    }
    po = Results(**results_dict)
    assert po.data.energy == -1.0


def test_compatibility_layer_for_stdout_on_results(prog_input_factory):
    """Test that the compatibility layer for stdout on Results works"""
    energy_input = prog_input_factory("energy")
    logs = "program standard out..."
    results_dict = {
        "input_data": energy_input,
        "success": True,
        "results": {"energy": -1.0},
        "provenance": {"program": "qcio-test-suite"},
        "stdout": logs,
    }
    results = Results(**results_dict)
    assert results.logs == logs



import json

from qcio import FileInput


def test_from_directory(test_data_dir):
    file_inp = FileInput.from_directory(
        directory=test_data_dir / "file_inputs",
        # program="terachem",
        extras={"my_extra": "123"},
    )

    # assert generic_inp.args.program == "terachem"
    assert file_inp.extras == {"my_extra": "123"}
    for filename, file in file_inp.files.items():
        data = (test_data_dir / "file_inputs" / filename).read_bytes()
        try:
            str_data = data.decode("utf-8")
        except UnicodeDecodeError:
            assert file == data
        else:
            assert file == str_data


def test_to_directory(test_data_dir, tmp_path):
    file_inp = FileInput.from_directory(
        directory=test_data_dir / "file_inputs",
        # program="terachem",
        extras={"my_extra": "123"},
    )
    file_inp.save_files(tmp_path)
    for filename in file_inp.files:
        # Ensure files were written
        assert (tmp_path / filename).exists()
        # Ensure files are identical to input files
        assert (tmp_path / filename).read_bytes() == (
            test_data_dir / "file_inputs" / filename
        ).read_bytes()


def test_to_from_file_with_binary_data(test_data_dir, tmp_path):
    file_inp = FileInput.from_directory(
        directory=test_data_dir / "file_inputs",
        # program="terachem",
        extras={"my_extra": "123"},
    )
    # Ensure that we have binary data
    assert isinstance(file_inp.files["c0"], bytes)
    # Write GenericInput to disk using .json() serialization
    file_inp.save(tmp_path / "file_input.json")
    # Ensure data was properly json serialized
    data = json.loads((tmp_path / "file_input.json").read_text())
    assert data["files"]["c0"].startswith("base64:")
    # Parse GenericInput from disk using deserialization
    file_inp_reloaded = FileInput.open(tmp_path / "file_input.json")
    # Assure all data returned to correct bytes or str representation
    assert file_inp_reloaded == file_inp



from pathlib import Path

import numpy as np
import pytest

from qcio import (
    DualProgramInput,
    FileInput,
    Files,
    OptimizationData,
    ProgramArgs,
    ProgramInput,
    Results,
    SinglePointData,
)
from qcio.utils import water as water_struct


@pytest.fixture
def test_data_dir():
    """Test data directory Path"""
    return Path(__file__).parent / "data"


@pytest.fixture
def water():
    """Water Structure fixture"""
    return water_struct


@pytest.fixture
def file_input():
    return FileInput(
        files={"binary": b"binary data", "text": "text data"},
        cmdline_args=["-i", "input.dat", "-o", "output.dat"],
    )


@pytest.fixture
def input_data(request, file_input, prog_input_factory, dprog_input_factory):
    """Input data fixture"""
    if request.param == "file_input":
        return file_input
    elif request.param == "calc_input":
        return prog_input_factory("energy")
    elif request.param == "ccalc_input":  # DualProgramInput
        return dprog_input_factory
    else:
        raise ValueError(f"Unknown input data type: {request.param}")


@pytest.fixture
def prog_input_factory(water):
    """Function that returns ProgramInput of calctype."""

    def _create_prog_inp(calctype):
        return ProgramInput(
            structure=water,
            calctype=calctype,
            model={"method": "hf", "basis": "sto-3g"},
            keywords={
                "maxiter": 100,
                "purify": "no",
                "some-bool": False,
                "displacement": 1e-3,
                "thermo_temp": 298.15,
            },
        )

    return _create_prog_inp


@pytest.fixture
def dprog_input_factory(water):
    """Function that returns DualProgramInput of calctype."""

    def _create_prog_inp(calctype):
        return DualProgramInput(
            structure=water,
            calctype=calctype,
            keywords={
                "maxiter": 100,
                "purify": "no",
                "some-bool": False,
            },
            subprogram="fake subprogram",
            subprogram_args=ProgramArgs(model={"method": "hf", "basis": "sto-3g"}),
        )

    return _create_prog_inp


@pytest.fixture
def sp_data():
    """SinglePointData object"""

    def _create_sp_data(structure):
        n_atoms = len(structure.symbols)
        gradient = np.arange(n_atoms * 3, dtype=np.float64).reshape(n_atoms, 3)
        hessian = np.arange(n_atoms**2 * 3**2, dtype=np.float64).reshape(
            n_atoms * 3, n_atoms * 3
        )
        return SinglePointData(
            energy=1.0,
            gradient=gradient,
            hessian=hessian,
            extras={"some_extra_result": 1},
        )

    return _create_sp_data


@pytest.fixture
def results(prog_input_factory, sp_data):
    """Successful Results object"""
    pi_energy = prog_input_factory("energy")
    sp_data = sp_data(pi_energy.structure)

    return Results[ProgramInput, SinglePointData](
        input_data=pi_energy,
        success=True,
        logs="program standard out...",
        data=sp_data,
        provenance={"program": "qcio-test-suite", "scratch_dir": "/tmp/qcio"},
        extras={"some_extra": 1},
    )


@pytest.fixture
def results_failure(prog_input_factory, sp_data):
    """Failed Results object"""
    pi_energy = prog_input_factory("energy")

    return Results[ProgramInput, Files](
        input_data=pi_energy,
        success=False,
        traceback="Traceback...",
        data=Files(),
        provenance={"program": "qcio-test-suite", "scratch_dir": "/tmp/qcio"},
        extras={"some_extra": 1},
    )


@pytest.fixture
def opt_data(results):
    return OptimizationData(trajectory=[results])



import numpy as np
import pytest

from qcio import Structure

from .data import structures as qcio_structures


def test_from_xyz(test_data_dir):
    xyz_str = (test_data_dir / "caffeine.xyz").read_text()
    caffeine = Structure.from_xyz(xyz_str)
    assert caffeine.symbols == qcio_structures.caffeine.symbols
    assert np.allclose(caffeine.geometry, qcio_structures.caffeine.geometry, rtol=1e-9)
    assert caffeine.multiplicity == qcio_structures.caffeine.multiplicity
    assert caffeine.charge == qcio_structures.caffeine.charge
    assert caffeine.identifiers.name == "caffeine"


def test_from_xyz_non_qcio_comments(test_data_dir):
    xyz_str = (test_data_dir / "caffeine.xyz").read_text()
    caffeine = Structure.from_xyz(xyz_str)
    assert caffeine.multiplicity == qcio_structures.caffeine.multiplicity
    assert caffeine.charge == qcio_structures.caffeine.charge
    assert caffeine.extras["xyz_comments"] == [
        "IUPAC",
        "1,3,7-trimethylpurine-2,6-dione",
    ]


def test_to_file_xyz(test_data_dir, tmp_path):
    caffeine = Structure.open(test_data_dir / "caffeine.xyz")
    caffeine.save(tmp_path / "caffeine_copy.xyz")
    caffeine_copy = Structure.open(tmp_path / "caffeine_copy.xyz")
    assert caffeine_copy.symbols == caffeine.symbols
    assert np.allclose(caffeine.geometry, qcio_structures.caffeine.geometry, rtol=1e-9)
    assert caffeine_copy.multiplicity == caffeine.multiplicity
    assert caffeine_copy.charge == caffeine.charge

    # single \n at end of file
    text = (tmp_path / "caffeine_copy.xyz").read_text()
    assert text.endswith("\n")
    assert not text.endswith("\n\n")


def test_to_xyz_comments(test_data_dir):
    caffeine = Structure.open(test_data_dir / "caffeine.xyz")
    xyz_str = caffeine.to_xyz()
    comments = xyz_str.split("\n")[1].split()
    assert "IUPAC" in comments
    assert "1,3,7-trimethylpurine-2,6-dione" in comments
    assert "qcio__identifiers_name=caffeine" in comments


def test_to_from_file_json(test_data_dir, tmp_path):
    caffeine = Structure.open(test_data_dir / "caffeine.xyz")
    caffeine.save(tmp_path / "caffeine_copy.json")
    caffeine_copy = Structure.open(tmp_path / "caffeine_copy.json")
    assert caffeine_copy.symbols == caffeine.symbols
    for i in range(len(caffeine.geometry)):
        assert caffeine_copy.geometry[i] == pytest.approx(
            caffeine.geometry[i], rel=1e-9
        )
    assert caffeine_copy.multiplicity == caffeine.multiplicity
    assert caffeine_copy.charge == caffeine.charge


def test_structure_model_dump_connectivity(water):
    # Test that connectivity is a list of lists of floats
    # Must cast all to the same type as toml cannot handle mixed types
    dict = water.model_dump()
    for bond in dict["connectivity"]:
        assert isinstance(bond, list)
        for val in bond:
            assert isinstance(val, float)


def test_structure_capitalize():
    structure = Structure(symbols=["NA"], geometry=[[0, 0, 0]])
    assert structure.symbols == ["Na"]
    structure = Structure(symbols=["h"], geometry=[[0, 0, 0]])
    assert structure.symbols == ["H"]


def test_atomic_symbols():
    structure = Structure(symbols=["Na", "Cl"], geometry=[[0, 0, 0], [1, 1, 1]])
    assert structure.atomic_numbers == [11, 17]


def test_ids_backwards_compatibility():
    struct = Structure(symbols=["H"], geometry=[[0, 0, 0]], ids={"name": "fake"})
    assert struct.identifiers.name == "fake"
    # Test that ids is an alias for identifiers
    assert struct.ids == struct.identifiers


@pytest.mark.parametrize(
    "xyzstring, charge, multiplicity, expected_charge, expected_multiplicity, raises",
    [
        (
            """2
            
            H 0.0 0.0 0.0
            H 0.0 0.0 1.0
            """,
            1,
            None,
            1,
            1,
            False,
        ),
        (
            """2
            
            H 0.0 0.0 0.0
            H 0.0 0.0 1.0
            """,
            None,
            2,
            0,
            2,
            False,
        ),
        (
            """2
            
            H 0.0 0.0 0.0
            H 0.0 0.0 1.0
            """,
            3,
            4,
            3,
            4,
            False,
        ),
        (
            """2
            qcio_charge=-3
            H 0.0 0.0 0.0
            H 0.0 0.0 1.0
            """,
            None,
            2,
            -3,
            2,
            False,
        ),
        (
            """2
            qcio_charge=-3
            H 0.0 0.0 0.0
            H 0.0 0.0 1.0
            """,
            1,
            None,
            None,
            None,
            True,
        ),
        (
            """2
            qcio_multiplicity=4
            H 0.0 0.0 0.0
            H 0.0 0.0 1.0
            """,
            1,
            None,
            1,
            4,
            False,
        ),
        (
            """2
            qcio_multiplicity=4
            H 0.0 0.0 0.0
            H 0.0 0.0 1.0
            """,
            None,
            1,
            None,
            None,
            True,
        ),
    ],
)
def test_passing_charge_and_multiplicity_to_open(
    tmp_path,
    xyzstring,
    charge,
    multiplicity,
    expected_charge,
    expected_multiplicity,
    raises,
):
    filepath = tmp_path / "test.xyz"
    filepath.write_text(xyzstring)

    if raises:
        with pytest.raises(ValueError):
            struct = Structure.open(filepath, charge=charge, multiplicity=multiplicity)
    else:
        struct = Structure.open(filepath, charge=charge, multiplicity=multiplicity)
        assert struct.charge == expected_charge
        assert struct.multiplicity == expected_multiplicity


def test_no_charge_multiplicity_to_non_xyz_files():
    with pytest.raises(ValueError):
        Structure.open("file.json", charge=1, multiplicity=2)


def test_multi_xyz(test_data_dir):
    xyz_str = (test_data_dir / "multi.xyz").read_text()
    structures = Structure.from_xyz_multi(xyz_str)
    assert len(structures) == 2

    # Capturing comments correctly
    assert structures[0].identifiers.name == "caffeine"
    assert structures[0].charge == -1
    assert structures[1].extras[Structure._xyz_comment_key] == ["-107.04437987"]

    # Add space between structures
    xyz_lines = xyz_str.split("\n")
    xyz_lines.insert(26, "\n")
    xyz_str_with_space = "\n".join(xyz_lines)
    structures_space = Structure.from_xyz_multi(xyz_str_with_space)
    assert len(structures_space) == 2
    assert structures_space[0].identifiers.name == "caffeine"
    assert structures_space[0].charge == -1
    assert structures_space[1].extras[Structure._xyz_comment_key] == ["-107.04437987"]

    assert structures[0] == structures_space[0]
    assert structures[1] == structures_space[1]

    # Make sure it works on a single structure
    caffeine = Structure.open(test_data_dir / "caffeine.xyz")
    assert caffeine.symbols == qcio_structures.caffeine.symbols
    assert np.allclose(caffeine.geometry, qcio_structures.caffeine.geometry, rtol=1e-9)
    assert caffeine.multiplicity == qcio_structures.caffeine.multiplicity
    assert caffeine.charge == qcio_structures.caffeine.charge
    assert caffeine.identifiers.name == "caffeine"


def test_distance():
    struct = Structure(symbols=["H", "H"], geometry=[[0, 0, 0], [0, 0, 1]])
    assert struct.distance(0, 1) == 1.0

    struct = Structure(symbols=["H", "H"], geometry=[[0, 0, 0], [0, 0, -1]])
    assert struct.distance(0, 1) == 1.0

    struct = Structure(symbols=["H", "H"], geometry=[[0, 0, 0], [0, 1.4, -1.3]])

    assert struct.distance(0, 1) == pytest.approx(1.91049731, abs=1e-8)


def test_reorder_indices():
    struct = Structure(
        symbols=["H", "O", "H"], geometry=[[1, 0, 0], [0, 0, 0], [0, 0, 1]]
    )
    struct.swap_indices([(0, 1), (1, 2), (2, 0)])

    assert struct.symbols == ["H", "H", "O"]
    assert np.array_equal(struct.geometry, [[0, 0, 1], [1, 0, 0], [0, 0, 0]])



from qcio import ProgramInput, Structure


def test_molecule_backwards_compatibility():
    structure = Structure(symbols=["Na", "Cl"], geometry=[[0, 0, 0], [1, 1, 1]])
    # Passing molecule=structure still works
    prog_input = ProgramInput(
        calctype="energy", molecule=structure, model={"method": "hf"}
    )
    assert prog_input.molecule == structure



import json

from qcio.models import Files


def test_binary_file(test_data_dir, tmp_path):
    input_filepath = test_data_dir / "file_inputs" / "c0"
    files = Files()
    files.add_file(input_filepath)
    assert isinstance(files.files[input_filepath.name], bytes)
    files.save_files(tmp_path)
    # Round trip of file is lossless
    assert (tmp_path / "c0").read_bytes() == input_filepath.read_bytes()


def test_str_file(test_data_dir, tmp_path):
    input_filepath = test_data_dir / "file_inputs" / "tc.in"
    files = Files()
    files.add_file(input_filepath)
    assert isinstance(files.files[input_filepath.name], str)
    files.save_files(tmp_path)
    # Round trip of file is lossless
    assert (tmp_path / "tc.in").read_text() == input_filepath.read_text()


def test_file_b64(test_data_dir):
    input_filepath = test_data_dir / "file_inputs" / "c0"
    files = Files()
    files.add_file(input_filepath)
    assert isinstance(files.files[input_filepath.name], bytes)
    json_str = files.model_dump_json()
    json_dict = json.loads(json_str)
    assert json_dict["files"][input_filepath.name].startswith("base64:")
    files_new = files.model_validate_json(json_str)
    # Round trip of file is lossless
    assert files_new.files["c0"] == input_filepath.read_bytes()


def test_dump_files(test_data_dir, tmp_path):
    data_dir = test_data_dir / "file_inputs"
    files = Files()
    files.add_file(data_dir / "c0")
    files.add_file(data_dir / "tc.in")
    files.save_files(tmp_path)

    assert (tmp_path / "c0").read_bytes() == (data_dir / "c0").read_bytes()
    assert (tmp_path / "tc.in").read_text() == (data_dir / "tc.in").read_text()


def test_files_recursive(test_data_dir, tmp_path):
    data_dir = test_data_dir / "file_inputs"
    files = Files()
    files.add_files(data_dir, recursive=True)
    files.save_files(tmp_path)

    for filename in files.files.keys():
        assert (tmp_path / filename).exists()
        assert (tmp_path / filename).read_bytes() == (data_dir / filename).read_bytes()



from qcio import OptimizationData, ProgramInput, Results
from qcio.view import generate_optimization_plot


def test_generate_optimization_plot_with_single_prog_output_failure(
    prog_input_factory, results_failure
):
    opt_input = prog_input_factory("optimization")

    prog_output = Results[ProgramInput, OptimizationData](
        input_data=opt_input,
        success=False,
        traceback="Traceback...",
        data=OptimizationData(trajectory=[results_failure]),
        provenance={"program": "fake-program"},
    )
    generate_optimization_plot(prog_output)



from qcio import Results


def test_serialization_to_disk_json(results, tmp_path):
    """Test serialization to disk json"""

    filename = tmp_path / "prog_output.json"
    results.save(filename)
    reopened = Results.open(filename)
    assert results == reopened

    # No filename or other extension to json by default
    filename = tmp_path / "prog_output.whatever"
    results.save(filename)
    reopened = Results.open(filename)
    assert results == reopened


def test_serialization_to_disk_yaml(results, tmp_path):
    """Test serialization to disk yaml"""
    for ext in [".yaml", ".yml"]:
        filename = tmp_path / f"prog_output{ext}"
        results.save(filename)
        reopened = Results.open(filename)
        assert results == reopened


def test_serialization_to_disk_toml(results, tmp_path):
    """Test serialization to disk toml"""

    filename = tmp_path / "prog_output.toml"
    results.save(filename)
    reopened = Results.open(filename)
    assert results == reopened



"""Utility functions for working with qcio objects."""

from __future__ import annotations

import json

import numpy as np
from pydantic import BaseModel

from .models import Structure

# Helper Structures
water = Structure(
    symbols=["O", "H", "H"],
    geometry=np.array(
        [
            [0.0253397, 0.01939466, -0.00696322],
            [0.22889176, 1.84438441, 0.16251426],
            [1.41760224, -0.62610794, -1.02954938],
        ]
    ),
    charge=0,
    multiplicity=1,
    connectivity=[(0, 1, 1.0), (0, 2, 1.0)],
    identifiers={"name": "water"},
)


def json_dumps(
    obj: BaseModel | list[BaseModel],
    exclude_unset: bool = True,
    **model_dump_kwargs,
) -> str:
    """Serialization helper for lists of pydantic objects.

    Args:
        obj: The object to serialize. Either a single pydantic object or a list of pydantic
            objects.
        exclude_unset: Whether to exclude unset fields from the serialized output.
        **model_dump_kwargs: Additional keyword arguments to pass to model_dump.
    """
    if isinstance(obj, list):
        return json.dumps(
            [
                o.model_dump(exclude_unset=exclude_unset, **model_dump_kwargs)
                for o in obj
            ]
        )
    return obj.model_dump_json(exclude_unset=exclude_unset, **model_dump_kwargs)



import importlib
import warnings
from importlib import metadata
from types import ModuleType
from typing import Any

from .models import *  # noqa: F403
from .models.utils import to_multi_xyz
from .utils import json_dumps

__version__ = metadata.version(__name__)


__all__ = [  # noqa: F405
    # Core Models
    "CalcType",
    "Model",
    "Provenance",
    "Structure",
    "Molecule",
    "Files",
    "FileInput",
    "ProgramInput",
    "ProgramOutput",
    "DualProgramInput",
    "SinglePointResults",
    "OptimizationResults",
    "Wavefunction",
    "ProgramArgs",
    "ProgramArgsSub",
    "json_dumps",
    "to_multi_xyz",
    "LengthUnit",
]

# Map of deprecated symbols → new fully-qualified path
_DEPRECATED: dict[str, str] = {
    "rmsd": "qcinf.rmsd",
    "align": "qcinf.align",
}


def __getattr__(name: str) -> Any:  # PEP 562
    """Redirect old qcio symbols to qcinf (with helpful message)."""
    if name not in _DEPRECATED:
        raise AttributeError(f"module 'qcio' has no attribute '{name}'")

    new_path = _DEPRECATED[name]
    mod_path, func_name = new_path.rsplit(".", 1)

    try:
        module: ModuleType = importlib.import_module(mod_path)
    except ModuleNotFoundError as exc:
        raise ModuleNotFoundError(
            f"`qcinf` is not installed, so `qcio.{name}` can no longer be "
            "used. Install the dependency with either:\n\n"
            "    python -m pip install qcinf\n"
            "or\n"
            "    python -m pip install 'qcinf[all]'\n\n"
            "and then import the new function location:\n\n"
            f"    from {mod_path} import {func_name}",
        ) from exc

    else:
        # Emit a FutureWarning so it's shown by default
        warnings.warn(
            f"`qcio.{name}` has moved to `{new_path}`.\n"
            "Please update your import to:\n\n"
            f"    from qcinf import {name}",
            FutureWarning,
            stacklevel=2,
        )
    return getattr(module, func_name)



from qcio import OptimizationData


def test_optimization_result_properties(results):
    opt_res = OptimizationData(
        trajectory=[results],
    )

    # Test properties
    assert opt_res.final_structure == results.input_data.structure
    assert opt_res.energies == [results.data.energy]
    assert opt_res.structures == [results.input_data.structure]
    # Test custom __repr_args__
    repr_args = opt_res.__repr_args__()
    assert isinstance(repr_args, list)
    for arg in repr_args:
        assert isinstance(arg, tuple)
        assert len(arg) == 2
        assert isinstance(arg[0], str)
        assert isinstance(arg[1], str)


def test_optimization_save_to_xyz(results, tmp_path):
    opt_res = OptimizationData(
        trajectory=[results] * 3,
    )
    opt_res.save(tmp_path / "opt_res.xyz")

    text = (tmp_path / "opt_res.xyz").read_text()
    # Text must be de-dented exactly as below
    correct_text = """3
qcio_charge=0 qcio_multiplicity=1 qcio__identifiers_name=water
O  0.01340919176202180 0.01026321207824930 -0.00368477733600419
H  0.12112430307330672 0.97600619725464122 0.08599884278042236
H  0.75016279902412597 -0.33132205318865016 -0.54481406902570462
3
qcio_charge=0 qcio_multiplicity=1 qcio__identifiers_name=water
O  0.01340919176202180 0.01026321207824930 -0.00368477733600419
H  0.12112430307330672 0.97600619725464122 0.08599884278042236
H  0.75016279902412597 -0.33132205318865016 -0.54481406902570462
3
qcio_charge=0 qcio_multiplicity=1 qcio__identifiers_name=water
O  0.01340919176202180 0.01026321207824930 -0.00368477733600419
H  0.12112430307330672 0.97600619725464122 0.08599884278042236
H  0.75016279902412597 -0.33132205318865016 -0.54481406902570462
"""
    assert text == correct_text


def test_optimization_save_non_xyz(results, tmp_path):
    opt_res = OptimizationData(
        trajectory=[results] * 3,
    )
    opt_res.save(tmp_path / "opt_res.json")
    opt_res_copy = OptimizationData.open(tmp_path / "opt_res.json")
    assert opt_res == opt_res_copy



"""End user results and data objects."""

from __future__ import annotations

import sys
import warnings
from itertools import product
from pathlib import Path
from typing import (
    Any,
    Generic,
    Literal,
    TypeVar,
    Union,
    get_args,
)

import numpy as np
from pydantic import BaseModel, ValidationInfo, field_validator, model_validator
from typing_extensions import Self

from qcio.helper_types import SerializableNDArray

from .base_models import CalcType, Files, Provenance, QCIOBaseModel
from .inputs import (
    FileInput,
    Inputs,
    ProgramInput,
)
from .inputs import (
    InputType as ProgramInputType,
)
from .structure import Structure
from .utils import deprecated_class, to_multi_xyz

__all__ = [
    "SinglePointData",
    "Wavefunction",
    "OptimizationData",
    "Results",
    "StructuredData",
    "StructuredDataType",
    "DataType",
    "Data",
    "ConformerSearchData",
    "SinglePointResults",
    "OptimizationResults",
    "ConformerSearchResults",
    "Results",
    "ProgramOutput",
]


class Wavefunction(QCIOBaseModel):
    """The wavefunction for a single point calculation.

    Attributes:
        scf_eigenvalues_a: The SCF alpha-spin orbital eigenvalues.
        scf_eigenvalues_b: The SCF beta-spin orbital eigenvalues.
        scf_occupations_a: The SCF alpha-spin orbital occupations.
        scf_occupations_b: The SCF beta-spin orbital occupations.
    """

    scf_eigenvalues_a: SerializableNDArray | None = None
    scf_eigenvalues_b: SerializableNDArray | None = None
    scf_occupations_a: SerializableNDArray | None = None
    scf_occupations_b: SerializableNDArray | None = None

    @field_validator(
        "scf_eigenvalues_a",
        "scf_eigenvalues_b",
        "scf_occupations_a",
        "scf_occupations_b",
    )
    @classmethod
    def to_numpy(cls, val, _info) -> np.ndarray | None:
        return np.asarray(val) if val is not None else None


class CalcInfoData(BaseModel):
    """Mixin for calcinfo attributes.

    Attributes:
        calcinfo_natoms: The number of atoms as computed by the program.
        calcinfo_nalpha: The number of alpha electrons as computed by the program.
        calcinfo_nbeta: The number of beta electrons as computed by the program.
        calcinfo_nbasis: The number of basis functions in the calculation.
        calcinfo_nmo: The number of molecular orbitals in the calculation.
    """

    # calcinfo contains general information about the calculation
    calcinfo_natoms: int | None = None
    calcinfo_nalpha: int | None = None
    calcinfo_nbeta: int | None = None
    calcinfo_nbasis: int | None = None
    calcinfo_nmo: int | None = None


class SinglePointData(Files, CalcInfoData):
    """The computed data from a single point calculation.

    Attributes:
        energy: The electronic energy of the structure in `Hartrees`.
        gradient: The gradient of the structure in `Hartrees/Bohr`.
        hessian: The hessian of the structure in `Hartrees/Bohr^2`.
        nuclear_repulsion_energy: The nuclear repulsion energy of the structure in
            Hartrees.

        wavefunction: Wavefunction data from the calculation.

        freqs_wavenumber: The frequencies of the structure in wavenumbers.
        normal_modes_cartesian: 3D n_vibmodes x n_atoms x 3 array containing
            un-mass-weighted Cartesian displacements of each normal mode in Bohr.
        gibbs_free_energy: Gibbs free energy (i.e. thermochemical analysis) in Hartrees
            of a system where translation / rotation / vibration degrees of freedom are
            approximated using ideal gas / rigid rotor / harmonic oscillator
            respectively.
        scf_dipole_moment: The x, y, z component of the dipole moment of the structure
            in units of e a0 (NOT Debye!).

    """

    # Core properties
    energy: float | None = None
    gradient: SerializableNDArray | None = None  # Coerced to 2D array
    hessian: SerializableNDArray | None = None  # Coerced to 2D array
    nuclear_repulsion_energy: float | None = None

    # Wavefunction data
    wavefunction: Wavefunction | None = None

    # Frequency data
    freqs_wavenumber: list[float] = []
    normal_modes_cartesian: SerializableNDArray | None = None  # Coerced to 3D array
    gibbs_free_energy: float | None = None

    # SCF results
    scf_dipole_moment: list[float] | None = None

    @field_validator("normal_modes_cartesian")
    @classmethod
    def _validate_normal_modes_cartesian_shape(
        cls, v: SerializableNDArray, info: ValidationInfo
    ):
        if v is not None:
            # Array must have length of the number of normal modes
            freqs = info.data.get("freqs_wavenumber")
            # Fallback to len(v) if we must
            n_normal_modes = len(freqs) if freqs else len(v)
            return np.asarray(v).reshape(n_normal_modes, -1, 3)
        return v

    @field_validator("gradient")
    @classmethod
    def _validate_gradient_shape(cls, v: SerializableNDArray):
        """Validate gradient is n x 3"""
        if v is not None:
            return np.asarray(v).reshape(-1, 3)

    @field_validator("hessian")
    @classmethod
    def _validate_hessian_shape(cls, v: SerializableNDArray):
        """Validate hessian is square"""
        if v is not None:
            v = np.asarray(v)
            n = int(np.sqrt(v.size))
            return v.reshape((n, n))

    def return_result(self, calctype: CalcType) -> float | SerializableNDArray:
        """Return the primary result of the calculation."""
        return getattr(self, calctype.value)

    @model_validator(mode="after")
    def _ensure_results(self) -> Self:
        """Ensure that at least one result is present."""
        if all(
            result is None
            for result in [
                self.energy,
                self.gradient,
                self.hessian,
            ]
        ):
            raise ValueError(
                "SinglePointResults requires either an energy, gradient, or hessian "
                "value."
            )
        return self


class OptimizationData(Files, CalcInfoData):
    """Computed data for an optimization (may be for a minimum or transition state).

    Attributes:
        energies: The energies for each step of the optimization.
        structures: The Structure objects for each step of the optimization.
        final_structure: The final, optimized Structure.
        trajectory: The Results objects for each step of the optimization.
    """

    trajectory: list[
        (Results[ProgramInput, SinglePointData] | Results[ProgramInput, Files])
    ] = []

    @property
    def final_structure(self) -> Structure:
        """The final Structure in the optimization."""
        return self.structures[-1]

    @property
    def final_molecule(self) -> Structure:
        warnings.warn(
            ".final_molecule is being deprecated and will be removed in a future. "
            "Please use .final_structure instead.",
            category=FutureWarning,
            stacklevel=2,
        )
        return self.final_structure

    @property
    def final_energy(self) -> float | None:  # Optional for np.nan
        """
        The final energy in the optimization. Is `np.nan` if final calculation failed.
        """
        return self.energies[-1]

    @property
    def energies(self) -> np.ndarray:
        """The energies for each step of the optimization."""
        return np.array(
            [
                # ensure_structured_results_on_success validator ensures .results
                # is not Files if success is True so # type: ignore
                output.data.energy if output.success else np.nan  # type: ignore
                for output in self.trajectory
            ],
            dtype=float,
        )

    @property
    def structures(self) -> list[Structure]:
        """The Structure objects for each step of the optimization."""
        return [output.input_data.structure for output in self.trajectory]

    def return_result(self, calctype: CalcType) -> Structure | None:
        """Return the primary result of the calculation."""
        return self.final_structure

    def __repr_args__(self):
        """Custom repr to avoid printing the entire collection of objects."""
        return [
            ("final_structure", f"{self.final_structure}"),
            ("trajectory", "[...]"),
            ("energies", "[...]"),
            ("structures", "[...]"),
        ]

    def to_xyz(self) -> str:
        """Return the trajectory as an `xyz` string."""
        return to_multi_xyz(
            prog_output.input_data.structure for prog_output in self.trajectory
        )

    def save(
        self,
        filepath: Path | str,
        exclude_none: bool = True,
        exclude_unset: bool = True,
        indent: int = 4,
        **kwargs: dict[str, Any],
    ) -> None:
        """Save an OptimizationOutput to a file.

        Args:
            filepath: The path to save the molecule to.
            exclude_none: If True, attributes with a value of None will not be written
                to the file.
            exclude_unset: If True, attributes that have not been set will not be
                written to the file.
            **kwargs: Additional keyword arguments to pass to the json serializer.

        Note:
            If the filepath has a `.xyz` extension, the trajectory will be saved to a
            multi-structure `xyz` file.
        """
        filepath = Path(filepath)
        if filepath.suffix == ".xyz":
            filepath.write_text(self.to_xyz())
            return
        super().save(filepath, exclude_none, exclude_unset, indent, **kwargs)


class ConformerSearchData(Files):
    """Data from a conformer search calculation.

    Conformers and rotamers are sorted by energy.

    Attributes:
        conformers: The conformers found in the search.
        conformer_energies: The energies for each conformer.
        rotamers: The rotamers found in the search.
        rotamer_energies: The energies for each rotamer.
    """

    conformers: list[Structure] = []
    conformer_energies: SerializableNDArray = np.array([])
    rotamers: list[Structure] = []
    rotamer_energies: SerializableNDArray = np.array([])

    @model_validator(mode="after")
    def _energies_size(self) -> ConformerSearchData:
        """Ensure the energies are the same size as the conformers and rotamers."""
        if self.conformer_energies.size > 0 and (
            self.conformer_energies.size != len(self.conformers)
        ):
            raise ValueError(
                "The number of conformer energies must match the number of conformers."
            )
        if self.rotamer_energies.size > 0 and (
            self.rotamer_energies.size != len(self.rotamers)
        ):
            raise ValueError(
                "The number of rotamer energies must match the number of rotamers."
            )
        return self

    @model_validator(mode="after")
    def _sort_by_energy(self) -> ConformerSearchData:
        """Sort conformers and rotamers by energy."""

        # Sort conformers and their energies together
        if self.conformer_energies.size > 0:
            sorted_indices = np.argsort(self.conformer_energies)
            self.conformers[:] = [self.conformers[i] for i in sorted_indices]
            self.conformer_energies[:] = self.conformer_energies[sorted_indices]

        # Sort rotamers and their energies together
        if self.rotamer_energies.size > 0:
            sorted_indices = np.argsort(self.rotamer_energies)
            self.rotamers[:] = [self.rotamers[i] for i in sorted_indices]
            self.rotamer_energies[:] = self.rotamer_energies[sorted_indices]

        return self

    @property
    def conformer_energies_relative(self) -> np.ndarray:
        """The relative energies for each conformer in the search."""
        if self.conformer_energies.size == 0:
            return np.array([])
        return self.conformer_energies - self.conformer_energies.min()

    @property
    def rotamer_energies_relative(self) -> np.ndarray:
        """The relative energies for each rotamer in the search."""
        if self.rotamer_energies.size == 0:
            return np.array([])
        return self.rotamer_energies - self.rotamer_energies.min()

    def conformers_filtered(
        self,
        threshold: float = 1.0,
        **rmsd_kwargs,
    ) -> tuple[list[Structure], SerializableNDArray]:
        """
        !!! warning "Moved since *qcio* 0.15.0"
            This convenience method has moved to
            [`qcinf.filter_conformers`][qcinf.filter_conformers]
            and this stub will be **removed** from *qcio* in a future release.

            ```python
            from qcinf import filter_conformers

            filtered_csr = filter_conformers(
                conformers=csr
                threshold=1.0,          # Bohr
                backend="qcinf",      # or "rdkit",
                **rmsd_kwargs,
            )
            ```
        """

        warnings.warn(
            "`ConformerSearchResults.conformers_filtered()` is deprecated. "
            "Install *qcinf* and use `qcinf.filter_conformers` instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        raise NotImplementedError(
            "Method removed.  Replace with:\n\n"
            "    from qcinf import filter_conformers\n\n"
            "    filtered_csr = filter_conformers(\n"
            "        prog_output.results,\n"
            "        threshold=1.0,\n"
            "        backend='qcinf',\n"
            "        **rmsd_kwargs\n"
            "    )"
        )


StructuredData = Union[SinglePointData, OptimizationData, ConformerSearchData]
StructuredDataType = TypeVar("StructuredDataType", bound=StructuredData)
Data = Union[Files, StructuredData]
DataType = TypeVar("DataType", bound=Data)


class Results(QCIOBaseModel, Generic[ProgramInputType, DataType]):
    """The core results object from a quantum chemistry calculation.

    Attributes:
        input_data: The input data for the calculation. Any of `qcio.Inputs`.
        success: Whether the calculation was successful.
        data: The data from the calculation. Contains parsed values and files.
            Any of `qcio.Data`.
        logs: The logs from the calculation.
        traceback: The traceback from the calculation, if it failed.
        provenance: The provenance information for the calculation.
        extras Dict[str, Any]: Additional information to bundle with the results. Use for
            schema development and scratch space.
        plogs str: `@property` Print the logs.
        ptraceback str: `@property` Print the traceback.
    """

    input_data: ProgramInputType
    success: Literal[True, False]
    data: DataType
    logs: str | None = None
    traceback: str | None = None
    provenance: Provenance

    @model_validator(mode="before")
    def _backwards_compatibility(cls, payload: dict[str, Any]) -> dict[str, Any]:
        """Backwards compatibility for renamed attributes."""

        # Backwards compatibility for .stdout attribute:
        if "stdout" in payload:
            warnings.warn(
                "The 'stdout' attribute has been renamed to 'logs'. Please update your "
                "code accordingly.",
                category=FutureWarning,
                stacklevel=2,
            )
            if "logs" not in payload:
                payload["logs"] = payload.pop("stdout")

        # Backwards compatibility for .results attribute:
        if "results" in payload:
            warnings.warn(
                "The 'results' attribute has been renamed to 'data'. Please update "
                "your code accordingly.",
                category=FutureWarning,
                stacklevel=2,
            )
            if isinstance(payload["results"], dict):
                payload["data"] = payload.pop("results")

        # Backwards compatibility for .files attribute:
        if "files" in payload:
            warnings.warn(
                "The 'files' attribute has been moved to 'data.files'. Please "
                "update your code accordingly.",
                category=FutureWarning,
                stacklevel=2,
            )
            # This moves files from the top level to the data attribute
            if isinstance(payload["data"], dict):
                data_files_dict = payload["data"].get("files", {})
            else:  # data["data"] is Files, SinglePointData, OptimizationData
                data_files_dict = payload["data"].files

            data_files_dict.update(**payload.pop("files"))

        return payload

    def model_post_init(self, __context) -> None:
        """Parameterize the class (if not set explicitly)."""
        # Check if the current class is still generic, do not override if explicitly set
        if self.__class__ is Results:
            input_type = type(self.input_data)
            results_type = type(self.data)
            self.__class__ = Results[input_type, results_type]  # type: ignore # noqa 501

    @property
    def results(self) -> DataType:
        """Return the data attribute."""
        warnings.warn(
            ".results has been renamed to .data. Please update your code accordingly.",
            category=FutureWarning,
            stacklevel=2,
        )
        return self.data

    @property
    def stdout(self) -> str | None:
        """Backwards compatibility for .stdout attribute."""
        warnings.warn(
            ".stdout has been renamed to .logs. Please update your code accordingly.",
            category=FutureWarning,
            stacklevel=2,
        )
        return self.logs

    @model_validator(mode="after")
    def ensure_traceback_on_failure(self) -> Self:
        if self.success is False:
            assert self.traceback is not None, (
                "A traceback must be provided for failed calculations."
            )
        return self

    @model_validator(mode="after")
    def _ensure_structured_results_on_success(self) -> Self:
        """Ensure structured results are provided for successful, non FileInputs."""
        # Covers case of ProgramInput and DualProgramInput
        if self.success is True and isinstance(self.input_data, ProgramInput):
            assert type(self.data) is not Files, (
                "Structured results must be provided for successful, non FileInput "
                "calculations."
            )

        return self

    @model_validator(mode="after")
    def ensure_primary_result_on_success(self) -> Self:
        if type(self.data) is SinglePointData:
            # Ensure the primary calctype result is present
            calctype_val = self.input_data.calctype.value  # type: ignore
            assert getattr(self.data, calctype_val) is not None, (
                f"Missing the primary result: {calctype_val}."
            )

        return self

    @property
    def plogs(self) -> None:
        """Print the logs"""
        print(self.logs)

    @property
    def pstdout(self) -> None:
        """Print the logs"""
        warnings.warn(
            ".pstdout has been renamed to .plogs. Please update your code accordingly.",
            category=FutureWarning,
            stacklevel=2,
        )
        print(self.logs)

    @property
    def ptraceback(self) -> None:
        """Print the traceback"""
        print(self.traceback)

    def __repr_args__(self) -> list[tuple[str, Any]]:
        """Exclude stdout and traceback from the repr and ensure success is first"""
        args = super().__repr_args__()

        # Replace stdout and traceback with "<...>"
        filtered_args = [
            (key, value if key not in {"stdout", "traceback"} else "<...>")
            for key, value in args
        ]
        # Ensure success is first
        success_arg = [(key, value) for key, value in filtered_args if key == "success"]
        other_args = [(key, value) for key, value in filtered_args if key != "success"]
        return success_arg + other_args

    @property
    def files(self) -> dict[str, str | bytes]:
        """Return the files attribute."""
        # Depreciation warning
        warnings.warn(
            ".files has been moved to .data.files. "
            "Please access it there going forward.",
            category=FutureWarning,
            stacklevel=2,
        )
        return self.data.files

    @property
    def return_result(self) -> float | SerializableNDArray | Structure | None:
        """Return the primary result of the calculation."""
        warnings.warn(
            ".return_result is being deprecated and will be removed in a future. "
            "Please access data directly at .data instead.",
            category=FutureWarning,
            stacklevel=2,
        )
        # For mypy
        assert self.data is not None, "No data exist on this Results object."
        assert type(self.input_data) is not FileInput, "FileInputs have no data."
        return self.data.return_result(self.input_data.calctype)  # type: ignore


### Backwards compatibility classes ###
@deprecated_class("Results")
class ProgramOutput(Results):
    """This class is deprecated and will be removed in a future release. Please use
    `Results` instead."""

    pass


@deprecated_class("SinglePointData")
class SinglePointResults(SinglePointData):
    """This class is deprecated and will be removed in a future release. Please use
    `SinglePointData` instead."""

    pass


@deprecated_class("OptimizationData")
class OptimizationResults(OptimizationData):
    """This class is deprecated and will be removed in a future release. Please use
    `OptimizationData` instead."""

    pass


@deprecated_class("ConformerSearchData")
class ConformerSearchResults(ConformerSearchData):
    """This class is deprecated and will be removed in a future release. Please use
    `ConformerSearchData` instead."""

    pass


### End backwards compatibility classes ###


# Register the concrete classes for serialization
def _register_program_output_classes():
    """Required so that pickle can find the concrete classes for serialization."""
    for spec_type, data_type in product(get_args(Inputs), get_args(Data)):
        # TODO: Remove ProgramOutput when compatibility is no longer needed
        for ClassType in [Results, ProgramOutput]:
            _class = ClassType[spec_type, data_type]
            name = _class.__name__
            this_module = sys.modules[__name__]
            if name not in this_module.__dict__:
                # Directly declare the type to ensure it is registered
                setattr(this_module, name, _class)


# Call this function during module initialization
_register_program_output_classes()



import copy

import numpy as np
import pytest
from pydantic import ValidationError

from qcio import ConformerSearchData


def test_energies_size(water):
    """Test that the number of energies matches the number of conformers"""
    w1 = water
    w2 = copy.deepcopy(water)

    # No energies is fine
    ConformerSearchData(
        conformers=[w1, w2],
    )

    # No energies is fine
    ConformerSearchData(
        rotamers=[w1, w2],
    )

    with pytest.raises(ValidationError):
        ConformerSearchData(
            conformers=[w1, w2],
            conformer_energies=[-1.0],
        )

    with pytest.raises(ValidationError):
        ConformerSearchData(
            rotamers=[w1, w2],
            rotamer_energies=[-1.0, -2.0, -3.0],
        )


def test_relative_energies(water):
    w1 = water
    w2 = copy.deepcopy(water)
    w3 = copy.deepcopy(water)

    # No energies returns empty array
    csr = ConformerSearchData(
        conformers=[w1, w2, w3],
    )

    np.testing.assert_array_equal(csr.conformer_energies_relative, np.array([]))

    # Energies are relative to the lowest energy
    csr = ConformerSearchData(
        conformers=[w1, w2, w3],
        conformer_energies=[-1.0, -2.0, -3.0],
    )
    np.testing.assert_array_equal(
        csr.conformer_energies_relative, np.array([0.0, 1.0, 2.0])
    )

    # No energies returns empty array
    csr = ConformerSearchData(
        rotamers=[w1, w2, w3],
    )

    np.testing.assert_array_equal(csr.conformer_energies_relative, np.array([]))

    # Energies are relative to the lowest energy
    csr = ConformerSearchData(
        rotamers=[w1, w2, w3],
        rotamer_energies=[-1.0, -2.0, -3.0],
    )
    np.testing.assert_array_equal(
        csr.rotamer_energies_relative, np.array([0.0, 1.0, 2.0])
    )


def test_conformer_search_results_sorting(water):
    w1 = water
    w2 = copy.deepcopy(water)
    w3 = copy.deepcopy(water)

    csr = ConformerSearchData(
        conformers=[w1, w2, w3],
        conformer_energies=[-1.0, -2.0, -3.0],
    )

    assert csr.conformers[0] is w3
    assert csr.conformers[1] is w2
    assert csr.conformers[2] is w1
    assert csr.conformer_energies.tolist() == [-3.0, -2.0, -1.0]




"""Utility functions for the models module."""

import warnings
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from qcio import Structure


def deprecated_function(new_name: str):
    """Notify users that a function is deprecated and will be removed in the future.
    Works equally well on functions and methods.

    Usage:
        @deprecated("new_name")
        def old_name():
            pass
    """

    def decorator(func):
        def wrapped(*args, **kwargs):
            warnings.warn(
                f"{func.__name__} is deprecated and will be removed in future "
                f"versions. Please use {new_name} instead.",
                FutureWarning,
                stacklevel=2,
            )
            return func(*args, **kwargs)

        return wrapped

    return decorator


def deprecated_class(new_name: str):
    """Notify users that a class is deprecated and will be removed in the future.
    Works only on Pydantic models since it relies on the model_post_init method.

    Usage:
        @deprecated_class("new_name")
        class OldName:
            pass
    """

    def class_decorator(cls):
        orig_init = cls.__init__

        def new_init(self, *args, **kwargs):
            warnings.warn(
                f"{cls.__name__} is deprecated and will be removed in a future "
                f"release. Please use '{new_name}' instead.",
                category=FutureWarning,
                stacklevel=2,
            )
            orig_init(self, *args, **kwargs)

        cls.__init__ = new_init
        return cls

    return class_decorator


def renamed_class(new_cls: Any):
    """Notify users that a class is deprecated and has been renamed.

    Usage:
        @renamed_class(NewClass)
        class OldClass:
            pass
    """

    def decorator(cls):
        def return_new_cls(cls, *args, **kwargs):
            warnings.warn(
                f"{cls.__name__} is deprecated and and will be removed in a future "
                f"release. Please use '{new_cls.__name__}' instead.",
                category=FutureWarning,
                stacklevel=2,
            )
            return new_cls(*args, **kwargs)

        cls.__new__ = return_new_cls
        return cls

    return decorator


def to_multi_xyz(structures: Iterable["Structure"]) -> str:
    """Create a multi-structure XYZ string from a list of structures.

    Args:
        structures: An Iterable of Structure objects.

    Returns:
        The multi-structure XYZ string.
    """
    return "".join(struct.to_xyz() for struct in structures)
